@@ -107,17 +107,15 @@ authldap_free_config removed - no longer required.
 #if HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
-#if HAVE_SYSLOG_H
-#include <syslog.h>
-#else
-#define syslog(a,b)
-#endif
 
 #include "authldap.h"
 #include "auth.h"
 #include "authldaprc.h"
+#include "debug.h"
 
+#ifndef DEBUG_LDAP
 #define DEBUG_LDAP 0
+#endif
 
 struct ldap_info
 {
@@ -129,15 +127,21 @@ struct ldap_info
 	const char *mail;
         const char *filter;
 	const char *domain;
+
 	uid_t uid;
 	gid_t gid;
 	int   timeout;
 	int   authbind;
 	int   deref;
+	int   protocol_version;
 	int   tls;
 
 	const char *mailroot;
 
+	char **auxoptions;
+	char **auxnames;
+	const char **attrlist;
+
 	/* Optional emailmap to handle */
 
 	const char *emailmap;
@@ -207,7 +211,7 @@ static int ldapconncheck()
 }
 
 static int read_env(const char *env, const char **copy,
-	const char *err, int needit, const char *value_default);
+	const char *errstr, int needit, const char *value_default);
 static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
 	char **copy, const char *username);
 
@@ -225,7 +229,7 @@ static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
  *   boolean
  */
 static int read_env(const char *env, const char **copy,
-	const char *err, int needit, const char *value_default)
+	const char *errstr, int needit, const char *value_default)
 {
 static char *ldapauth=0;
 static size_t ldapauth_size=0;
@@ -284,8 +288,7 @@ int	l=strlen(env);
 
 	if (needit)
 	{
-		fprintf(stderr, "ERR: %s\n",err);
-		fflush(stderr);
+		err("%s", errstr);
 		return 0;
 	}
 
@@ -306,139 +309,301 @@ int	l=strlen(env);
  */
 static int authldap_read_config(struct ldap_info *ldap)
 {
-  struct passwd *pwent;
-  struct group  *grent;
-  const char *p;
+	struct passwd *pwent;
+	struct group  *grent;
+	const char *p;
+	size_t i, pass;
 
-  memset(ldap,0,sizeof(struct ldap_info));
+	for (i=0; ldap->auxoptions && ldap->auxoptions[i]; i++)
+		free(ldap->auxoptions[i]);
+	for (i=0; ldap->auxnames && ldap->auxnames[i]; i++)
+		free(ldap->auxnames[i]);
+
+	if (ldap->attrlist)
+		free(ldap->attrlist);
+	if (ldap->auxnames)
+		free(ldap->auxnames);
+	if (ldap->auxoptions)
+		free(ldap->auxoptions);
+
+	memset(ldap,0,sizeof(struct ldap_info));
 	
-  if (!read_env("LDAP_SERVER",&ldap->hostname,"You need to specify a ldap server in config file",1,NULL))
-    return 0;
+	if (!read_env("LDAP_SERVER",&ldap->hostname,
+		      "You need to specify a ldap server in config file",1,NULL))
+		return 0;
 
-  if (!read_env("LDAP_AUTHBIND", &p, "", 0, ""))
-	return (0);
+	if (!read_env("LDAP_AUTHBIND", &p, "", 0, ""))
+		return (0);
 
-  if (p)
-    sscanf(p,"%d",&ldap->authbind);
+	if (p)
+		sscanf(p,"%d",&ldap->authbind);
 
-  ldap->port=LDAP_PORT;
+	ldap->port=LDAP_PORT;
 
-  if (!read_env("LDAP_PORT", &p, "", 0, ""))
-	return (0);
+	if (!read_env("LDAP_PORT", &p, "", 0, ""))
+		return (0);
 
-  if (p)
-    sscanf(p,"%d",&ldap->port);
+	if (p)
+		sscanf(p,"%d",&ldap->port);
 
-  if (!read_env("LDAP_BASEDN",&ldap->basedn,"You need to specify a basedn in config file",1,NULL))
-    return 0;
-  if (!read_env("LDAP_BINDDN",&ldap->binddn,"You need to specify a BINDDN in config file",0,NULL))
-    return 0;
-  if (!read_env("LDAP_BINDPW",&ldap->bindpw,"You need to specify a password for the BINDDN in config file",0,NULL))
-    return 0;
-  if (!read_env("LDAP_MAIL",&ldap->mail,"You need to specify a attribute for mail in config file",0,"mail"))
-    return 0;
-  if (!read_env("LDAP_DOMAIN",&ldap->domain,"You need to specify a domain for mail in config file",0,""))
-    return 0;
+	if (!read_env("LDAP_BASEDN",&ldap->basedn,
+		      "You need to specify a basedn in config file",1,NULL))
+		return 0;
+	if (!read_env("LDAP_BINDDN",&ldap->binddn,
+		      "You need to specify a BINDDN in config file",0,NULL))
+		return 0;
+	if (!read_env("LDAP_BINDPW",&ldap->bindpw,
+		      "You need to specify a password for the BINDDN in config file",0,NULL))
+		return 0;
+	if (!read_env("LDAP_MAIL",&ldap->mail,
+		      "You need to specify a attribute for mail in config file",0,"mail"))
+		return 0;
+	if (!read_env("LDAP_DOMAIN",&ldap->domain,
+		      "You need to specify a domain for mail in config file",0,""))
+		return 0;
 
-  p=0;
-  ldap->uid=0;
-  if (!read_env("LDAP_GLOB_UID", &p, "", 0, ""))
-	return (0);
+	p=0;
+	ldap->uid=0;
+	if (!read_env("LDAP_GLOB_UID", &p, "", 0, ""))
+		return (0);
 
-  if (p && *p)
-  {
-  unsigned long n;
-
-	if (sscanf(p, "%lu", &n) == 1)
-		ldap->uid=(uid_t)n;
-	else
+	if (p && *p)
 	{
-		pwent=getpwnam(p);
-		if (!pwent)
+		unsigned long n;
+
+		if (sscanf(p, "%lu", &n) == 1)
+			ldap->uid=(uid_t)n;
+		else
 		{
-			syslog(LOG_DAEMON|LOG_CRIT,
-				"authldap: INVALID LDAP_GLOB_UID\n");
-			return (0);
+			pwent=getpwnam(p);
+			if (!pwent)
+			{
+				err("authldap: INVALID LDAP_GLOB_UID");
+				return (0);
+			}
+			ldap->uid=pwent->pw_uid;
 		}
-		ldap->uid=pwent->pw_uid;
 	}
-  }
 
-  ldap->gid=0;
-  p=0;
-  if (!read_env("LDAP_GLOB_GID", &p, "", 0, ""))
-	return (0);
+	ldap->gid=0;
+	p=0;
+	if (!read_env("LDAP_GLOB_GID", &p, "", 0, ""))
+		return (0);
 
-  if (p && *p)
-  {
-  unsigned long n;
-
-	if (sscanf(p, "%lu", &n) == 1)
-		ldap->gid=(gid_t)n;
-	else
+	if (p && *p)
 	{
-		grent=getgrnam(p);
-		if (!grent)
+		unsigned long n;
+
+		if (sscanf(p, "%lu", &n) == 1)
+			ldap->gid=(gid_t)n;
+		else
 		{
-			syslog(LOG_DAEMON|LOG_CRIT,
-				"authldap: INVALID LDAP_GLOB_GID\n");
-			return (0);
+			grent=getgrnam(p);
+			if (!grent)
+			{
+				err("authldap: INVALID LDAP_GLOB_GID");
+				return (0);
+			}
+			ldap->gid=grent->gr_gid;
 		}
-		ldap->gid=grent->gr_gid;
 	}
-  }
 
-  ldap->timeout=5;
-  p=0;
-  if (read_env("LDAP_TIMEOUT", &p, "", 0, "") && p)
-  {
-    sscanf(p,"%d",&ldap->timeout);
-  }
+	ldap->timeout=5;
+	p=0;
+	if (read_env("LDAP_TIMEOUT", &p, "", 0, "") && p)
+	{
+		sscanf(p,"%d",&ldap->timeout);
+	}
 
-  ldap->tls=0;
-  p=0;
-  if (read_env("LDAP_TLS", &p, "", 0, "") && p)
-  {
-    ldap->tls=atoi(p);
-  }
+	ldap->tls=0;
+	p=0;
+	if (read_env("LDAP_TLS", &p, "", 0, "") && p)
+	{
+		ldap->tls=atoi(p);
+	}
 
-  ldap->filter=0;
-  p=0;
-  if (read_env("LDAP_FILTER", &p, "", 0, "") && p && strlen (p))
-  {
-    ldap->filter=p;
-  }
+	ldap->filter=0;
+	p=0;
+	if (read_env("LDAP_FILTER", &p, "", 0, "") && p && strlen (p))
+	{
+		ldap->filter=p;
+	}
 
-  ldap->deref = LDAP_DEREF_NEVER; 
-  p=0;
-  if (!read_env("LDAP_DEREF", &p, "", 0, ""))
-	return (0);
-  if (p)
-  {
-	if (!strcasecmp (p, "never"))
-	  ldap->deref = LDAP_DEREF_NEVER;
-	else if (!strcasecmp (p, "searching"))
-	  ldap->deref = LDAP_DEREF_SEARCHING;
-	else if (!strcasecmp (p, "finding"))
-	  ldap->deref = LDAP_DEREF_FINDING;
-	else if (!strcasecmp (p, "always"))
-	  ldap->deref = LDAP_DEREF_ALWAYS; 
-  }
+	ldap->deref = LDAP_DEREF_NEVER; 
+	ldap->protocol_version = 0;	/* use API default */
+	p=0;
+	if (!read_env("LDAP_DEREF", &p, "", 0, ""))
+		return (0);
+	if (p)
+	{
+#ifndef LDAP_OPT_DEREF
+		err("authldaplib: LDAP_OPT_DEREF not available, ignored");
+#endif
+		if (!strcasecmp (p, "never"))
+			ldap->deref = LDAP_DEREF_NEVER;
+		else if (!strcasecmp (p, "searching"))
+			ldap->deref = LDAP_DEREF_SEARCHING;
+		else if (!strcasecmp (p, "finding"))
+			ldap->deref = LDAP_DEREF_FINDING;
+		else if (!strcasecmp (p, "always"))
+			ldap->deref = LDAP_DEREF_ALWAYS; 
+	}
 
-  if (!read_env("LDAP_MAILROOT",&ldap->mailroot,"",0,NULL)
-      || ldap->mailroot == NULL || ldap->mailroot[0] == 0)
-	  ldap->mailroot=NULL;
+	if (!read_env("LDAP_PROTOCOL_VERSION", &p, 0, 0, 0))
+		return (0);
+	if (p)
+	{
+	int lpv = atoi(p);
+#ifndef LDAP_OPT_PROTOCOL_VERSION
+		err("authldaplib: LDAP_OPT_PROTOCOL_VERSION not available, ignored");
+#endif
+		if (lpv == 0
+#ifdef LDAP_VERSION_MIN
+			|| lpv < LDAP_VERSION_MIN
+#endif
+#ifdef LDAP_VERSION_MAX
+			|| lpv > LDAP_VERSION_MAX
+#endif
+		)
+			err("authldaplib: illegal protocol version ignored");
+		else
+			ldap->protocol_version = lpv;
+	}
 
-  if (!read_env("LDAP_EMAILMAP", &ldap->emailmap, "", 0, "") ||
-      !read_env("LDAP_EMAILMAP_BASEDN", &ldap->emailmap_basedn, "", 0, "") ||
-      !read_env("LDAP_EMAILMAP_ATTRIBUTE", &ldap->emailmap_handle, "", 0, "")||
-      !read_env("LDAP_EMAILMAP_MAIL",
-		&ldap->emailmap_handle_lookup, "", 0, ""))
-	return (0);
+	if (!read_env("LDAP_MAILROOT",&ldap->mailroot,"",0,NULL)
+	    || ldap->mailroot == NULL || ldap->mailroot[0] == 0)
+		ldap->mailroot=NULL;
 
-  return 1;
+	if (!read_env("LDAP_EMAILMAP", &ldap->emailmap, "", 0, "") ||
+	    !read_env("LDAP_EMAILMAP_BASEDN", &ldap->emailmap_basedn, "", 0, "") ||
+	    !read_env("LDAP_EMAILMAP_ATTRIBUTE", &ldap->emailmap_handle, "", 0, "")||
+	    !read_env("LDAP_EMAILMAP_MAIL",
+		      &ldap->emailmap_handle_lookup, "", 0, ""))
+		return (0);
+
+
+	for (pass=0; pass<2; pass++)
+	{
+		if (pass)
+		{
+			if ((ldap->auxnames=malloc((i+1)*sizeof(char *)))
+			    == NULL ||
+			    (ldap->auxoptions=malloc((i+1)*sizeof(char *)))
+			    == NULL)
+			{
+				perror("CRIT: authldap: malloc failed");
+				if (ldap->auxnames)
+					ldap->auxnames[0]=0;
+				return 0;
+			}
+		}
+		i=0;
+
+		if (pass)
+		{
+			ldap->auxnames[0]=NULL;
+			ldap->auxoptions[0]=NULL;
+		}
+
+		if (!read_env("LDAP_AUXOPTIONS", &p, "", 0, NULL)
+		    || p == NULL || *p == 0)
+			p=NULL;
+
+		while (p && *p)
+		{
+			size_t n;
+
+			if (*p == ',')
+			{
+				++p;
+				continue;
+			}
+
+			for (n=0; p[n] && p[n] != ',' && p[n] != '='; n++)
+				;
+
+			if (pass)
+			{
+				if ((ldap->auxoptions[i]=malloc(n+1)) == NULL)
+				{
+					perror("CRIT: authldap: malloc failed");
+					return 0;
+				}
+
+				memcpy(ldap->auxoptions[i], p, n);
+				ldap->auxoptions[i][n]=0;
+				ldap->auxoptions[i+1]=NULL;
+			}
+
+			p += n;
+
+			if (*p == '=') ++p;
+
+			for (n=0; p[n] && p[n] != ','; n++)
+				;
+
+			if (pass)
+			{
+				if (n == 0)
+				{
+					if ((ldap->auxnames[i]=
+					     strdup(ldap->auxoptions[i]))
+					    == NULL)
+					{
+						perror("CRIT: authldap: malloc failed");
+						return 0;
+					}
+				}
+				else if ((ldap->auxnames[i]=malloc(n+1)) == NULL)
+				{
+					perror("CRIT: authldap: malloc failed");
+					return 0;
+				}
+				else
+				{
+					memcpy(ldap->auxnames[i], p, n);
+					ldap->auxnames[i][n]=0;
+					ldap->auxnames[i+1]=NULL;
+				}
+			}
+			p += n;
+			++i;
+		}
+	}
+
+	if ((ldap->attrlist=malloc((i+20)*sizeof(const char *))) == NULL)
+	{
+		perror("CRIT: authldap: malloc failed");
+		return 0;
+	}
+
+	return 1;
 }
 
+static char **get_values(LDAP *ld, LDAPMessage *entry, const char *attribut)
+{
+	char ** values;
+	values=ldap_get_values(ld,entry, (char *)attribut);
+
+	if (values==NULL)
+	{
+#ifdef HAVE_LDAP_RESULT2ERROR
+		int ld_errno = ldap_result2error(ld,entry,0);
+		if (ld_errno && ld_errno != LDAP_DECODING_ERROR)
+			/* We didn't ask for this attribute */
+			ldap_perror(ld,"ldap_get_values");
+#else
+		if (ld->ld_errno != LDAP_DECODING_ERROR)
+			/* We didn't ask for this attribute */
+			ldap_perror(ld,"ldap_get_values");
+#endif
+	}
+
+	return values;
+}
+
+
+
 /*
  * Function: copy_value
  *   Copy value from a LDAP attribute to $copy
@@ -475,8 +640,7 @@ static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
 	else if (ldap_count_values(values)>1)
 	 {
 		 *copy=strdup(values[0]);
-		 syslog(LOG_DAEMON,
-			"authldaplib: duplicate attribute %s for %s\n",
+		 fprintf(stderr, "WARN: authldaplib: duplicate attribute %s for %s\n",
 			attribut,
 			username);
 		 *copy=NULL;
@@ -491,7 +655,7 @@ static void copy_value(LDAP *ld, LDAPMessage *entry, const char *attribut,
 		 *copy=strdup(values[0]);
 	 }
 #if DEBUG_LDAP
-  syslog(LOG_DAEMON|LOG_CRIT,"copy_value %s: %s\n",attribut,values[0]);
+  dprintf("copy_value %s: %s",attribut,values[0]);
 #endif
   ldap_value_free(values);
 }
@@ -538,10 +702,7 @@ static int enable_tls_on(LDAP *conn) {
 	{
 		const char *s=ldap_err2string(ldrc);
 
-#if	HAVE_SYSLOG_H
-		syslog(LOG_DAEMON|LOG_CRIT,
-		       "ldap_get_option failed: %s", s);
-#endif
+		err("ldap_get_option failed: %s", s);
 		return (-1);
 	}
 
@@ -558,14 +719,12 @@ static int enable_tls_on(LDAP *conn) {
 	{
 		const char *s=ldap_err2string(ldrc);
 
-#if	HAVE_SYSLOG_H
-		syslog(LOG_DAEMON|LOG_CRIT,
-		       "ldap_start_tls_s failed: %s", s);
-#endif
+		err("ldap_start_tls_s failed: %s", s);
 		return (-1);
 	}
 	return 0;
 #else
+	err("authldaplib: TLS not available");
 	return (-1);
 #endif
 }
@@ -575,31 +734,27 @@ static LDAP *ldapconnect()
 LDAP	*p;
 
 #if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"Hostname: %s:%d\n",my_ldap.hostname,my_ldap.port);
-	syslog(LOG_DAEMON|LOG_CRIT,"UID:      %d\n",my_ldap.uid);
-	syslog(LOG_DAEMON|LOG_CRIT,"GID:      %d\n",my_ldap.gid);
+	dprintf("Hostname: %s:%d",my_ldap.hostname,my_ldap.port);
+	dprintf("UID:      %d",my_ldap.uid);
+	dprintf("GID:      %d",my_ldap.gid);
 #endif
 
 	if (ldapconncheck())
+	{
+		dprintf("authldaplib: timing out after failed connection");
 		return (NULL);
+	}
 
 	p=ldap_init((char *)my_ldap.hostname,my_ldap.port);
 
 	if (p==NULL)
 	{
-#if	HAVE_SYSLOG_H
-		syslog(LOG_DAEMON|LOG_CRIT,
-			"Cannot connect to LDAP server (%s:%d): %s",
+		err("cannot connect to LDAP server (%s:%d): %s",
 			my_ldap.hostname, my_ldap.port, strerror(errno));
-#else
-		fprintf(stderr,
-			"Cannot connect to LDAP server (%s:%d): %s",
-			my_ldap.hostname, my_ldap.port, strerror(errno));
-#endif
 		ldapconnfailure();
 	}
 #if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"ldapconnect end\n");
+	dprintf("ldapconnect end");
 #endif
 	return (p);
 }
@@ -611,7 +766,10 @@ int     ldrc;
 	if (my_ldap_fp)	return (0);
 
 	if (authldap_read_config(&my_ldap) == 0)
+	{
+		err("authldaplib: error in LDAP configuration file, aborting");
 		return (1);
+	}
 
 	my_ldap_fp=ldapconnect();
 
@@ -620,26 +778,41 @@ int     ldrc;
 		return (1);
 	}
 
-#if HAVE_LDAP_TLS
+#ifdef LDAP_OPT_PROTOCOL_VERSION
+
+	/* Set protocol version if selected */
+	if (my_ldap.protocol_version &&
+		ldaperror(ldrc = ldap_set_option(my_ldap_fp, LDAP_OPT_PROTOCOL_VERSION,
+		(void *) & my_ldap.protocol_version)) != LDAP_SUCCESS)
+	  {
+		const char *s=ldap_err2string(ldrc);
+
+		err("ldap_set_option(PROTOCOL_VERSION %d) failed: %s",
+			my_ldap.protocol_version, s);
+		authldapclose();
+		ldapconnfailure();
+		return (-1);
+	  }
+	if (my_ldap.protocol_version)
+		dprintf("selected ldap protocol version %d", my_ldap.protocol_version);
+#endif
+
 	if (my_ldap.tls && enable_tls_on(my_ldap_fp))
 	{
 		authldapclose();
 		ldapconnfailure();
 		return (-1);
 	}
-#endif
 
 #ifdef LDAP_OPT_DEREF
 
-	/* Set deferencing mode */
+	/* Set dereferencing mode */
 	if (ldaperror(ldrc = ldap_set_option(my_ldap_fp, LDAP_OPT_DEREF,
 					 (void *) & my_ldap.deref)) != LDAP_SUCCESS)
 	  {
 		const char *s=ldap_err2string(ldrc);
 
-#if	HAVE_SYSLOG_H
-		syslog(LOG_DAEMON|LOG_CRIT, "ldap_set_option failed: %s", s);
-#endif
+		err("ldap_set_option(DEREF) failed: %s", s);
 		authldapclose();
 		ldapconnfailure();
 		return (-1);
@@ -647,9 +820,13 @@ int     ldrc;
 #endif
 
   /* Bind to server */
-#if DEBUG_LDAP
-  syslog(LOG_DAEMON|LOG_CRIT,"BindDn:   %s\nBindPw:   %s\n",my_ldap.binddn,my_ldap.bindpw);
-#endif
+  if (auth_debug_login_level >= 2)
+	  dprintf("binding to LDAP server as DN '%s', password '%s'",
+	  	my_ldap.binddn ? my_ldap.binddn : "<null>",
+	  	my_ldap.bindpw ? my_ldap.bindpw : "<null>");
+  else
+	  dprintf("binding to LDAP server as DN '%s'",
+	  	my_ldap.binddn ? my_ldap.binddn : "<null>");
 
   if (ldaperror(ldrc = ldap_simple_bind_s(my_ldap_fp,
 		(char *)my_ldap.binddn,
@@ -657,9 +834,7 @@ int     ldrc;
     {
     const char *s=ldap_err2string(ldrc);
 
-#if	HAVE_SYSLOG_H
-	syslog(LOG_DAEMON|LOG_CRIT, "ldap_simple_bind_s failed: %s", s);
-#endif
+	err("ldap_simple_bind_s failed: %s", s);
 	authldapclose();
 	ldapconnfailure();
 	return (-1);
@@ -760,9 +935,10 @@ static int auth_ldap_do2(const char *service,
 
 	if (!srch)
 	{
-		perror("malloc");
+		perror("CRIT: authldaplib: malloc");
 		exit(1);
 	}
+	dprintf("using emailmap search: %s", srch);
 
 	tv.tv_sec=my_ldap.timeout;
 	tv.tv_usec=0;
@@ -784,6 +960,7 @@ static int auth_ldap_do2(const char *service,
 	{
 		free(srch);
 
+		dprintf("ldap_search_st failed");
 		if (my_ldap_fp)	return (-1);
 		return (1);
 	}
@@ -793,8 +970,7 @@ static int auth_ldap_do2(const char *service,
 		free(srch);
 
 		if (cnt)
-			syslog(LOG_DAEMON|LOG_CRIT,
-			       "authldap: %d results from search: %s\n",
+			err("emailmap: %d entries returned from search %s (but we need exactly 1)",
 			       cnt, srch);
 		ldap_msgfree(result);
 		return -1;
@@ -807,8 +983,7 @@ static int auth_ldap_do2(const char *service,
 	{
 		ldap_msgfree(result);
 
-		syslog(LOG_DAEMON|LOG_CRIT,
-		       "authldap: unexpected NULL from ldap_first_entry");
+		err("authldap: unexpected NULL from ldap_first_entry");
 		return -1;
 	}
 
@@ -816,6 +991,7 @@ static int auth_ldap_do2(const char *service,
 
 	if (!v)
 	{
+		dprintf("emailmap: empty attribute");
 		ldap_msgfree(result);
 		return (-1);
 	}
@@ -824,6 +1000,8 @@ static int auth_ldap_do2(const char *service,
 	if (aname[0] == 0)
 		aname=my_ldap.mail;
 
+	dprintf("emailmap results: aname=%s, handle=%s", aname, v);
+
 	cnt=auth_ldap_do3(service,
 			  aname, v, pass, callback, arg, newpass, user);
 
@@ -840,8 +1018,7 @@ static int auth_ldap_do3(const char *service,
 			 const char *authaddr)
 {
 	char *newpass_crypt=0;
-	const char *attributes[10], *ldap_attributes[10];
-  
+	const char *attributes[10];
 	struct timeval timeout;
 
 	LDAPMessage *result;
@@ -854,6 +1031,7 @@ static int auth_ldap_do3(const char *service,
 	char *mailDir=0;
 	char *userPassword=0;
 	char *cryptPassword=0;
+	char *options=0;
 	char *cn=0;
 	uid_t au;
 	gid_t ag;
@@ -900,6 +1078,8 @@ static int auth_ldap_do3(const char *service,
             strcat(filter, ")");
         }
 
+	dprintf("using search filter: %s", filter);
+
 	timeout.tv_sec=my_ldap.timeout;
 	timeout.tv_usec=0;
 
@@ -919,17 +1099,21 @@ static int auth_ldap_do3(const char *service,
 	for (i=0; i<9; i++)
 	{
 		if (attributes[i])
-			ldap_attributes[j++]=attributes[i];
+			my_ldap.attrlist[j++]=attributes[i];
 	}
 
-	ldap_attributes[j]=0;
+	for (i=0; my_ldap.auxoptions[i]; i++)
+		my_ldap.attrlist[j++]=my_ldap.auxoptions[i];
+
+	my_ldap.attrlist[j]=0;
 
 	if (ldaperror(ldap_search_st(my_ldap_fp,
 				     (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
-				     filter, (char **)ldap_attributes, 0,
+				     filter, (char **)my_ldap.attrlist, 0,
 				     &timeout, &result)
 		      != LDAP_SUCCESS))
 	{
+		dprintf("ldap_search_st() failed");
 		free(filter);
 
 		if (my_ldap_fp)	return (-1);
@@ -941,18 +1125,15 @@ static int auth_ldap_do3(const char *service,
 	/* If we are more than one result, reject */
 	if (ldap_count_entries(my_ldap_fp,result)!=1)
 	{
+		dprintf("number of entries returned: %d (but we need exactly 1)",
+			ldap_count_entries(my_ldap_fp,result));
 		ldap_msgfree(result);
 		return -1;
 	}
-#if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"Nombre de rsulat:    %d\n",ldap_count_entries(my_ldap_fp,result));
-#endif
 
 	dn = ldap_get_dn(my_ldap_fp, result);
 
-#if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"DN:    %s\n",dn);
-#endif
+	dprintf("one entry returned, DN: %s", dn ? dn : "<null>");
 
 	if (dn == NULL) 
 	{
@@ -970,8 +1151,39 @@ static int auth_ldap_do3(const char *service,
 	}
 
 #if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"after ldap_first_entry\n");
+	dprintf("after ldap_first_entry");
 #endif
+
+	/* print all the raw attributes */
+	if (auth_debug_login_level >= 2)
+	{
+	char *attr;
+	BerElement *berptr = 0;
+
+		attr = ldap_first_attribute(my_ldap_fp, entry, &berptr);
+		if (attr) dprintf("raw ldap entry returned:");
+		while (attr)
+		{
+		char **av, **ap;
+		
+			av = ldap_get_values(my_ldap_fp, entry, attr);
+			ap = av;
+			if (av)
+			{
+				while(*ap)
+				{
+					dprintf("| %s: %s", attr, *ap);
+					ap++;
+				}
+				ldap_value_free(av);
+			}
+			ldap_memfree(attr);
+			attr = ldap_next_attribute(my_ldap_fp, entry, berptr);
+		}
+		
+		ber_free(berptr, 0);
+	}
+
 	/* Copy the directory and the password into struct */
 	copy_value(my_ldap_fp,entry,attributes[0],&homeDir, user);
 	if (attributes[1])
@@ -996,7 +1208,7 @@ static int auth_ldap_do3(const char *service,
 			free(p);
 		}
 #if DEBUG_LDAP
-		syslog(LOG_DAEMON|LOG_CRIT,"au= %d\n",au);
+		dprintf("au= %d",au);
 #endif
 	}
 
@@ -1012,7 +1224,7 @@ static int auth_ldap_do3(const char *service,
 			free(p);
 		}
 #if DEBUG_LDAP
-		syslog(LOG_DAEMON|LOG_CRIT,"ag= %d\n",ag);
+		dprintf("ag= %d",ag);
 #endif
 	}
 
@@ -1026,7 +1238,7 @@ static int auth_ldap_do3(const char *service,
 
 		if (!new_mailroot)
 		{
-			syslog(LOG_DAEMON|LOG_CRIT, "authldap: malloc failed");
+			perror("CRIT: authldap: malloc failed");
 			rc= -1;
 		}
 		else
@@ -1038,6 +1250,55 @@ static int auth_ldap_do3(const char *service,
 		}
 	}
 
+	j=1;
+
+	for (i=0; my_ldap.auxoptions[i]; i++)
+	{
+		char *val;
+
+		copy_value(my_ldap_fp, entry, my_ldap.auxoptions[i], &val,
+			   user);
+
+		if (!val)
+			continue;
+
+		j += 2 + strlen(my_ldap.auxnames[i]) +
+			strlen(val);
+		free(val);
+	}
+
+	options=malloc(j);
+
+	if (!options)
+	{
+		perror("CRIT: authldap: malloc failed");
+		rc= -1;
+	}
+	else
+	{
+		*options=0;
+
+		for (i=0; my_ldap.auxoptions[i]; i++)
+		{
+			char *val;
+
+			copy_value(my_ldap_fp, entry, my_ldap.auxoptions[i],
+				   &val,
+				   user);
+
+			if (!val)
+				continue;
+
+			if (*options)
+				strcat(options, ",");
+			strcat(options, my_ldap.auxnames[i]);
+			strcat(options, "=");
+			strcat(options, val);
+			free(val);
+		}
+	}
+
+
 	auth.sysusername=user;
 	auth.sysuserid= &au;
 	auth.sysgroupid= ag;
@@ -1048,6 +1309,7 @@ static int auth_ldap_do3(const char *service,
 	auth.clearpasswd=userPassword;
 	auth.passwd=cryptPassword;
 	auth.quota=quota;
+	auth.options=options && *options ? options:NULL;
 
 	if (auth.sysusername == 0)
 		auth.sysusername=auth.address="";
@@ -1059,44 +1321,64 @@ static int auth_ldap_do3(const char *service,
 
 	if (au == 0 || ag == 0)
 	{
-		syslog(LOG_DAEMON|LOG_CRIT,
-		       "authlib: refuse to authenticate %s: uid=%d, gid=%d\n",
+		err("authldaplib: refuse to authenticate %s: uid=%d, gid=%d (zero uid or gid not permitted)",
 		       user, au, ag);
 		rc= 1;
 	}
 
+	auth_debug_authinfo("DEBUG: authldaplib: ", &auth,
+		userPassword, cryptPassword);
+
 	if (pass)
 	{
 		if (my_ldap.authbind) 
 		{
+			/* FIXME: if LDAP_V3 is available we should use a persistent connection for this */
 			LDAP *bindp=ldapconnect();
 
 			if (!bindp)
+			{
+				dprintf("ldapconnect failed");
 				rc=1;
+			}
 			else
 			{
-#if HAVE_LDAP_TLS
-				if(my_ldap.tls && enable_tls_on(bindp)) {
-#if HAVE_SYSLOG_H
-					syslog(LOG_DAEMON|LOG_CRIT, "authlib: LDAP_TLS enabled but I'm unable to start tls, check your config\n");
+#ifdef LDAP_OPT_PROTOCOL_VERSION
+				/* Set protocol version */
+				if (my_ldap.protocol_version &&
+					ldap_set_option(bindp, LDAP_OPT_PROTOCOL_VERSION,
+					(void *) & my_ldap.protocol_version) != LDAP_OPT_SUCCESS)
+				  {
+					err("ldap_set_option(PROTOCOL_VERSION %d) failed",
+						my_ldap.protocol_version);
+					rc = 1;
+				  }
+				else
 #endif
+
+				if(my_ldap.tls && enable_tls_on(bindp)) {
+					err("authldaplib: LDAP_TLS enabled but I'm unable to start tls, check your config");
 					rc = 1;
 				} else {
-#endif
-					switch (ldap_simple_bind_s(bindp, dn, (char *)pass))
+				int ldrc;
+					dprintf("rebinding with DN '%s' to validate password", dn);
+					ldrc = ldap_simple_bind_s(bindp, dn, pass);
+					switch (ldrc)
 					{
 					case LDAP_SUCCESS:
+						dprintf("authentication bind successful");
 						break;
 					case LDAP_INVALID_CREDENTIALS:
+						dprintf("authentication bind failed, invalid credentials");
 						rc = -1;
 						break;
 					default:
+						dprintf("authentication bind failed, some other problem: %s",
+							ldap_err2string(ldrc));
 						rc = 1;
 						break;
 					}
-#if HAVE_LDAP_TLS
 				}
-#endif
 				ldap_unbind(bindp);
 			}
 			if (rc == 0 && newpass)
@@ -1112,7 +1394,14 @@ static int auth_ldap_do3(const char *service,
 			if (auth.clearpasswd)
 			{
 				if (strcmp(pass,auth.clearpasswd))
+				{
+					if (auth_debug_login_level >= 2)
+						dprintf("supplied password '%s' does not match clearpasswd '%s'",
+							pass, auth.clearpasswd);
+					else
+						dprintf("supplied password does not match clearpasswd");
 					rc= -1;
+				}
 			}
 			else
 			{
@@ -1121,7 +1410,12 @@ static int auth_ldap_do3(const char *service,
 				if (p && strncasecmp(p, "{crypt}", 7) == 0)
 					p += 7; /* For authcheckpassword */
 
-				if (!p || authcheckpassword(pass, p))
+				if (!p)
+				{
+					dprintf("no password to compare against!");
+					rc= -1;
+				}
+				else if (authcheckpassword(pass, p))
 					rc= -1;
 			}
 
@@ -1183,17 +1477,17 @@ static int auth_ldap_do3(const char *service,
 		free(newpass_crypt);
 	free (dn);
 #if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"before callback rc=%d\n",rc);
+	dprintf("before callback rc=%d",rc);
 #endif
 
 	if (rc == 0 && callback)
 		rc= (*callback)(&auth, arg);
 #if DEBUG_LDAP
-	syslog(LOG_DAEMON|LOG_CRIT,"after callback rc=%d\n",rc);
+	dprintf("after callback rc=%d",rc);
 #endif
 
 	ldap_msgfree(result);
-
+	if (options)	free(options);
 	if (homeDir)	free(homeDir);
 	if (mailDir)	free(mailDir);
 	if (userPassword)	free(userPassword);
@@ -1342,3 +1636,140 @@ static char *var_expand(const char *str, const struct varlist *vl)
 
 	return (q);
 }
+
+void auth_ldap_enumerate( void(*cb_func)(const char *name,
+					 uid_t uid,
+					 gid_t gid,
+					 const char *homedir,
+					 const char *maildir,
+					 void *void_arg),
+			   void *void_arg)
+{
+	char *filter;
+	const char *attributes[5];
+	const char *ldap_attributes[6];
+
+	LDAPMessage *result;
+	LDAPMessage *entry;
+	int n, i, j;
+	struct timeval timeout;
+
+	if (ldapopen())	return;
+
+	if (my_ldap.filter)
+	{
+		filter=strdup(my_ldap.filter);
+	}
+	else
+	{
+		filter=malloc(strlen(my_ldap.mail)+4);
+		if (filter)
+			strcat(strcpy(filter, my_ldap.mail), "=*");
+
+	}
+
+	if (!filter)
+	{
+		perror("malloc");
+		return;
+	}
+
+	read_env("LDAP_MAIL", &attributes[0], "", 0, "mail");
+	read_env("LDAP_UID", &attributes[1], "", 0, 0);
+	read_env("LDAP_GID", &attributes[2], "", 0, 0);
+	read_env("LDAP_HOMEDIR", &attributes[3], "", 0, "homeDir");
+	read_env("LDAP_MAILDIR", &attributes[4], "", 0, 0);
+
+	j=0;
+	for (i=0; i<5; i++)
+	{
+		if (attributes[i])
+			ldap_attributes[j++]=attributes[i];
+	}
+
+	ldap_attributes[j]=0;
+
+	timeout.tv_sec=my_ldap.timeout;
+	timeout.tv_usec=0;
+
+	if (ldaperror(ldap_search_st(my_ldap_fp,
+				     (char *)my_ldap.basedn,LDAP_SCOPE_SUBTREE,
+				     filter, (char **)ldap_attributes, 0,
+				     &timeout, &result)
+		      != LDAP_SUCCESS))
+	{
+		free(filter);
+		return;
+	}
+
+	n=ldap_count_entries(my_ldap_fp,result);
+
+	for (i=0; i<n; i++)
+	{
+		char **names;
+		int n;
+
+		entry=i ? ldap_next_entry(my_ldap_fp, result):
+			ldap_first_entry(my_ldap_fp, result);
+
+		if (entry == NULL)
+			break;
+
+		names=get_values(my_ldap_fp, entry, attributes[0]);
+		n=ldap_count_values(names);
+		for (i=0; i<n; i++)
+		{
+			const char *name=names[i];
+
+			char *uid_s=NULL;
+			char *gid_s=NULL;
+			char *homedir;
+			char *maildir;
+			uid_t uid=my_ldap.uid;
+			gid_t gid=my_ldap.gid;
+
+			if (attributes[1])
+			{
+				copy_value(my_ldap_fp, entry, attributes[1],
+					   &uid_s,
+					   name);
+			}
+
+			if (attributes[2])
+			{
+				copy_value(my_ldap_fp, entry, attributes[1],
+					   &gid_s, name);
+			}
+
+			copy_value(my_ldap_fp, entry, attributes[3],
+				   &homedir, name);
+			copy_value(my_ldap_fp, entry, attributes[4],
+				   &maildir, name);
+
+			if (uid_s)
+				uid=atol(uid_s);
+			if (gid_s)
+				gid=atol(gid_s);
+
+			if (name && homedir)
+				(*cb_func)(name, uid, gid, homedir,
+					   maildir, void_arg);
+
+
+			if (uid_s)
+				free(uid_s);
+			if (gid_s)
+				free(gid_s);
+			if (homedir)
+				free(homedir);
+			if (maildir)
+				free(maildir);
+		}
+		ldap_value_free(names);
+	}
+
+	ldap_msgfree(result);
+
+	if (i == n)
+		(*cb_func)(NULL, 0, 0, NULL, NULL, void_arg);
+}
